"""
consistent_trees_hdf5 frontend



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.frontends.consistent_trees_hdf5.arbor import \
    ConsistentTreesHDF5Arbor
